/*
 * $Id: comm.c,v 1.4 2001/10/23 09:30:46 kr Exp $
 */
#define _KERNEL

#include "headers.h"
#include "papillon.h"
#include "refcnt.h"
#include "papctl.h"

extern pap_config_t config;
extern krwlock_t config_rwlock;

int64_t pap_comm(pap_config_t * c, enum request r)
{
    int64_t ret = 0;

    inc_refcnt();
    switch (r) {

    case SET:
	cmn_err(CE_CONT, "SET request for Papillon configuration.\n");

	/*
	 * This is the only case where we lock the config. Therefore
	 * we are usign a rwlock and not a mutex.
	 */
	if (!rw_tryenter(&config_rwlock, RW_WRITER)) {
	    ret = CONFIG_BLOCKED;
	    break;
	}

	if (copyin(c, &config, sizeof(config)))
	    ret = MEM_FAULT;

	rw_exit(&config_rwlock);


	/*
	 * hides module if requested in config
	 */
	hide_module();
	break;

    case GET:
	cmn_err(CE_CONT, "GET request for Papillon configuration.\n");

	rw_enter(&config_rwlock, RW_READER);

	if (copyout(&config, c, sizeof(config)))
	    ret = MEM_FAULT;

	rw_exit(&config_rwlock);
	break;

    default:
	ret = set_errno(EINVAL);
    }

    dec_refcnt();
    return ret;
}
