/*
 * Papillon - Solaris security module -  http://www.roqe.org/papillon
 * Copyright (c) 2000,2001 Konrad Rieck <kr@roqe.org>
 * $Id: papillon.c,v 1.26 2001/11/02 15:24:58 kr Exp $
 *
 * The main part of the Papillon module. 
 */

#define _KERNEL

#include "headers.h"
#include "papillon.h"
#include "refcnt.h"

/*
 * Runtime configuration
 */
pap_config_t config = {
    /* rstproc, ppromisc, modhiding, secstdfd */
    ON, ON, ON, ON,

    /* fifoprot, symprot, hardprot */
    DENY, DENY, DENY
};

/*
 * Hidden files
 */
pap_modfiles_t modfiles[] = {
    {"modctl->mod_filename", NULL, NULL},
    {INITSCRIPT, NULL, NULL},
    {RC2SCRIPT, NULL, NULL},
    {CONTROLPATH, NULL, NULL},
    {NULL, NULL, NULL}
};

/*
 * That's me. My modctl struct, my vnode and my parent vnode. 
 */
struct modctl *this_module;

/*
 * This is the loadable module wrapper.
 */
extern struct mod_ops mod_miscops;

/*
 * Structure of the system-entry table.
 */
#ifdef _64BIT
extern struct sysent sysent32[];
#else
extern struct sysent sysent[];
#endif


/* 
 * Structure of the loaded modules
 */
extern struct modctl modules;

/*
 * Module linkage information for the kernel.
 */
static struct modlmisc modlmisc = {
    &mod_miscops,
    "Papillon v" VERSION
};

struct modlinkage modlinkage = {
    MODREV_1,
    (void *) &modlmisc,
    NULL
};

/*
 * Syscall redirection vector table. 
 */
pap_syscall_t syscalls[] = {

#if defined(HARDPROT)
    {0, (int64_t(*)())pap_link, SE_NOUNLOAD | SE_LOADED, SYS_link,
     DFLT_ARG, 0},
#else
    {0, 0, SE_NOUNLOAD | SE_LOADED, SYS_UNUSED, DFLT_ARG, 0},
#endif

#if defined(SYMPROT) || defined(FIFOPROT)
    {0, (int64_t(*)())pap_open, SE_NOUNLOAD | SE_LOADED, SYS_open,
     DFLT_ARG, 0},
    {0, (int64_t(*)())pap_open64, SE_NOUNLOAD | SE_LOADED, SYS_open64,
     DFLT_ARG, 0},
#else
    {0, 0, SE_NOUNLOAD | SE_LOADED, SYS_UNUSED, DFLT_ARG, 0},
    {0, 0, SE_NOUNLOAD | SE_LOADED, SYS_UNUSED, DFLT_ARG, 0},
#endif

#if defined(SECSTDFD)
    {0, (int64_t(*)())pap_execve, SE_NOUNLOAD | SE_LOADED, SYS_execve,
     DFLT_ARG, 0},
#else
    {0, 0, SE_NOUNLOAD | SE_LOADED, SYS_UNUSED, DFLT_ARG, 0},
#endif

#if defined(PPROMISC)
    {0, (int64_t(*)())pap_putmsg, SE_NOUNLOAD | SE_LOADED, SYS_putmsg,
     DFLT_ARG, 0},
#else
    {0, 0, SE_NOUNLOAD | SE_LOADED, SYS_UNUSED, DFLT_ARG, 0},
#endif
    {0, (int64_t(*)())pap_comm, SE_NOUNLOAD | SE_LOADED, SYS_papcomm, 2, 0},
    {0, 0, 0, 0, 0, 0}
};

#ifdef _64BIT
/*
 * Syscall redirection vector table. 
 */
pap_syscall_t syscalls64[] = {

#if defined(HARDPROT)
    {0, (int64_t(*)())pap_link_64, SE_NOUNLOAD | SE_LOADED, SYS_link,
     DFLT_ARG, 0},
#else
    {0, 0, SE_NOUNLOAD | SE_LOADED, SYS_UNUSED, DFLT_ARG, 0},
#endif

#if defined(SYMPROT) || defined(FIFOPROT)
    {0, (int64_t(*)())pap_open_64, SE_NOUNLOAD | SE_LOADED, SYS_open,
     DFLT_ARG, 0},
    {0, (int64_t(*)())pap_open64_64, SE_NOUNLOAD | SE_LOADED, SYS_open64,
     DFLT_ARG, 0},
#else
    {0, 0, SE_NOUNLOAD | SE_LOADED, SYS_UNUSED, DFLT_ARG, 0},
    {0, 0, SE_NOUNLOAD | SE_LOADED, SYS_UNUSED, DFLT_ARG, 0},
#endif

#if defined(SECSTDFD)
    {0, (int64_t(*)())pap_execve_64, SE_NOUNLOAD | SE_LOADED, SYS_execve,
     DFLT_ARG, 0},
#else
    {0, 0, SE_NOUNLOAD | SE_LOADED, SYS_UNUSED, DFLT_ARG, 0},
#endif

#if defined(PPROMISC)
    {0, (int64_t(*)())pap_putmsg_64, SE_NOUNLOAD | SE_LOADED, SYS_putmsg,
     DFLT_ARG, 0},
#else
    {0, 0, SE_NOUNLOAD | SE_LOADED, SYS_UNUSED, DFLT_ARG, 0},
#endif
    {0, (int64_t(*)())pap_comm, SE_NOUNLOAD | SE_LOADED, SYS_papcomm, 2, 0},
    {0, 0, 0, 0, 0, 0}
};

#endif


/*
 * Pointer to the proc filesystem vnode operations. 
 */
static struct vnodeops *procfs_vnodeops;

/*
 * Backup pointer for the old praccess(), readdir() and lookup() functions.
 */
int (*old_praccess) ();
int (*old_vnreaddir) ();
int (*old_vnlookup) ();

/*
 * Mutexes, Locks
 */
kmutex_t promisc_lock;
kmutex_t readdir_lock;
kmutex_t refcnt_lock;
krwlock_t config_rwlock;

/*
 * Restore the old syscall vectors saved in the syscall struct.
 */
void restore_old_syscalls()
{
    int i;

    for (i = 0; syscalls[i].index; i++) {
	if (syscalls[i].index != SYS_UNUSED) {
	    dcmn_err((CE_CONT, "Restoring syscall %d.\n",
		      syscalls[i].index));
#ifdef _64BIT
	    sysent32[syscalls[i].index].sy_callc = syscalls[i].sc;
	    sysent32[syscalls[i].index].sy_flags = syscalls[i].flags;

	    if (syscalls[i].narg != DFLT_ARG)
		sysent32[syscalls[i].index].sy_narg = syscalls[i].onarg;

	    sysent[syscalls64[i].index].sy_callc = syscalls64[i].sc;
	    sysent[syscalls64[i].index].sy_flags = syscalls64[i].flags;

	    if (syscalls64[i].narg != DFLT_ARG)
		sysent[syscalls64[i].index].sy_narg = syscalls64[i].onarg;

#else
	    sysent[syscalls[i].index].sy_callc = syscalls[i].sc;
	    sysent[syscalls[i].index].sy_flags = syscalls[i].flags;

	    if (syscalls[i].narg != DFLT_ARG)
		sysent[syscalls[i].index].sy_narg = syscalls[i].onarg;
#endif
	}
    }

}

/*
 * Store the old syscall vectors in the syscall struct.
 */
void store_old_syscalls()
{
    int i;

    for (i = 0; syscalls[i].index; i++) {
	if (syscalls[i].index != SYS_UNUSED) {
	    dcmn_err((CE_CONT, "Storing syscall %d.\n", syscalls[i].index));
#ifdef _64BIT
	    syscalls[i].sc =
		(int64_t(*)())sysent32[syscalls[i].index].sy_callc;
	    syscalls[i].flags = sysent32[syscalls[i].index].sy_flags;
	    syscalls[i].onarg = sysent32[syscalls[i].index].sy_narg;

	    syscalls64[i].sc =
		(int64_t(*)())sysent[syscalls64[i].index].sy_callc;
	    syscalls64[i].flags = sysent[syscalls64[i].index].sy_flags;
	    syscalls64[i].onarg = sysent[syscalls64[i].index].sy_narg;

#else
	    syscalls[i].sc =
		(int64_t(*)())sysent[syscalls[i].index].sy_callc;
	    syscalls[i].flags = sysent[syscalls[i].index].sy_flags;
	    syscalls[i].onarg = sysent[syscalls[i].index].sy_narg;
#endif
	}
    }
}

/*
 * Set the new syscall vectors specified in the syscall struct.
 */
void set_new_syscalls()
{
    int i;

    for (i = 0; syscalls[i].index; i++) {
	if (syscalls[i].index != SYS_UNUSED) {
	    dcmn_err((CE_CONT, "Setting new syscall %d.\n",
		      syscalls[i].index));
#ifdef _64BIT
	    sysent32[syscalls[i].index].sy_callc = syscalls[i].nc;
	    sysent32[syscalls[i].index].sy_flags = syscalls[i].flags;

	    if (syscalls[i].narg != DFLT_ARG)
		sysent32[syscalls[i].index].sy_narg = syscalls[i].narg;

	    sysent[syscalls64[i].index].sy_callc = syscalls64[i].nc;
	    sysent[syscalls64[i].index].sy_flags = syscalls64[i].flags;

	    if (syscalls64[i].narg != DFLT_ARG)
		sysent[syscalls64[i].index].sy_narg = syscalls64[i].narg;

#else
	    sysent[syscalls[i].index].sy_callc = syscalls[i].nc;
	    sysent[syscalls[i].index].sy_flags = syscalls[i].flags;

	    if (syscalls[i].narg != DFLT_ARG)
		sysent[syscalls[i].index].sy_narg = syscalls[i].narg;
#endif
	}
    }
}

#ifdef MODHIDING
/*
 * Lookup vnodes for files to hide. 
 */
void lookup_vnodes()
{
    int i;

    if (!this_module)
	cmn_err(CE_PANIC, "Could not find module Papillon in the loaded "
		"module list. But I AM module Papillon !?");

    /*
     * This is myself. Keep this in mind.
     */
    modfiles[0].path = this_module->mod_filename;

    for (i = 0; modfiles[i].path; i++)
	lookupname(modfiles[i].path, UIO_SYSSPACE, NO_FOLLOW,
		   &modfiles[i].parent_vnode, &modfiles[i].vnode);

    if (!modfiles[0].vnode || !modfiles[0].parent_vnode)
	cmn_err(CE_PANIC, "Could not find the module's file. This is "
		"impossible! Papillon has to be loaded from a file.");
}

/*
 * Release vnodes.. 
 */
void release_vnodes()
{
    int i;

    for (i = 0; modfiles[i].path; i++) {
	if (modfiles[i].vnode)
	    VN_RELE(modfiles[i].vnode);
	if (modfiles[i].parent_vnode)
	    VN_RELE(modfiles[i].parent_vnode);
    }

}

#endif

/*
 * Init and load the module.
 */
int _init()
{
    int i, cri;
    struct modctl *this;

    procfs_vnodeops = (struct vnodeops *) modgetsymvalue("prvnodeops", 0);

    if (!procfs_vnodeops) {
	cmn_err(CE_WARN, "Symbol prvnodeops not found, procfs module not "
		" loaded.");
	return set_errno(ENOENT);
    }

    if ((i = mod_install(&modlinkage)) != 0) {
	cmn_err(CE_WARN, "Could not load module.");
	return i;
    } else {
	cmn_err(CE_CONT, "Module successfully loaded.\n");
    }

    /*
     * Initialize mutexes
     */
    dcmn_err((CE_CONT, "Initializing locks.\n"));
    mutex_init(&refcnt_lock, NULL, MUTEX_DEFAULT, NULL);
    mutex_init(&promisc_lock, NULL, MUTEX_DEFAULT, NULL);
    mutex_init(&readdir_lock, NULL, MUTEX_DEFAULT, NULL);
    rw_init(&config_rwlock, NULL, RW_DRIVER, NULL);

    /*
     * Module is loaded, find it in the modules list
     */
    this = &modules;
    while (this->mod_linkage != &modlinkage) {
	this = this->mod_next;
	ASSERT(this != &modules);
    }
    this_module = this;

    /*
     * Look up file vnode and parent vnode of the modules filename
     * and store them for later use in hiding.c
     */
#if defined(MODHIDING)
    lookup_vnodes();
#endif

    /*
     * Initialize reference counter
     */
    init_refcnt();

    dcmn_err((CE_CONT, "Entering critical area.\n"));
    cri = ddi_enter_critical();

    /*
     * Change vnode operation for the filename of the module and its
     * parent vnode. 
     */
#if defined(MODHIDING)
    if (modfiles[0].vnode->v_op) {
	old_vnlookup = modfiles[0].vnode->v_op->vop_lookup;
	old_vnreaddir = modfiles[0].vnode->v_op->vop_readdir;
	modfiles[0].vnode->v_op->vop_lookup = pap_vnlookup;
	modfiles[0].vnode->v_op->vop_readdir = pap_vnreaddir;
    }

    hide_module();
#endif

    /*
     * Change proc vnode access function. 
     */
#if defined(RSTPROC)
    if (procfs_vnodeops) {
	old_praccess = procfs_vnodeops->vop_access;
	procfs_vnodeops->vop_access = pap_praccess;
    }
#endif

    /*
     * Redirect syscalls. 
     */
    store_old_syscalls();
    set_new_syscalls();

    ddi_exit_critical(cri);
    dcmn_err((CE_CONT, "Leaving critical area.\n"));

    return i;
}

int _info(struct modinfo *modinfop)
{
    return (mod_info(&modlinkage, modinfop));
}

/* 
 * "fini" and unload the module.
 */
int _fini()
{
    int i, cri;

    if (!check_refcnt())
	return set_errno(EBUSY);

    dcmn_err((CE_CONT, "Entering critical area.\n"));
    cri = ddi_enter_critical();

    /*
     * Restore old syscall vectors. 
     */
    restore_old_syscalls();

    /*
     * Restore old vnode operations.
     */
#if defined(MODHIDING)
    if (modfiles[0].vnode->v_op) {
	modfiles[0].vnode->v_op->vop_lookup = old_vnlookup;
	modfiles[0].vnode->v_op->vop_readdir = old_vnreaddir;
    }
#endif

#if defined(RSTPROC)
    if (procfs_vnodeops)
	procfs_vnodeops->vop_access = old_praccess;
#endif

    ddi_exit_critical(cri);
    dcmn_err((CE_CONT, "Leaving critical area.\n"));

#if defined(MODHIDING)
    release_vnodes();
#endif

    /*
     * Destroy mutexes.
     */
    dcmn_err((CE_CONT, "Destroying locks.\n"));
    rw_destroy(&config_rwlock);
    mutex_destroy(&readdir_lock);
    mutex_destroy(&promisc_lock);
    mutex_destroy(&refcnt_lock);

    if ((i = mod_remove(&modlinkage)) != 0)
	cmn_err(CE_WARN, "Could not unload module.");
    else
	cmn_err(CE_CONT, "Module successfully unloaded.\n");

    return i;
}
