/*
 * $Id: suid.c,v 1.7 2001/11/02 14:06:56 kr Exp $
 */
#define _KERNEL

#include "headers.h"
#include "papillon.h"
#include "refcnt.h"

#if defined(SECSTDFD)

#define STDIO(i) i==0?"STDIN":i==1?"STDOUT":"STDERR"

extern pap_syscall_t syscalls[];
#ifdef _64BIT
extern pap_syscall_t syscalls64[];
#endif

extern pap_config_t config;
extern krwlock_t config_rwlock;

int suid_check(char *path)
{
    vnode_t *vp;
    vattr_t vattr;
    boolean_t ret = B_FALSE;

    if (lookupname((char *) path, UIO_USERSPACE, NO_FOLLOW, NULL, &vp))
	return B_FALSE;

    if (VOP_GETATTR(vp, &vattr, 0, CRED()))
	goto err_and_release;

    if (vattr.va_mode & (S_ISUID | S_ISGID))
	ret = B_TRUE;

  err_and_release:
    VN_RELE(vp);
    return ret;
}

int64_t pap_execve(const char *path, char *const argv[], char *const envp[])
{
    uf_info_t *ufinfo;
    short i;
    boolean_t suid = B_FALSE, fake_opened_fd[3];
    char k_path[256];
    int64_t ret;

    inc_refcnt();
    dcmn_err((CE_CONT, "execve() syscall.\n"));
    rw_enter(&config_rwlock, RW_READER);

    if (!config.secstdfd)
	goto skip;

    suid = suid_check((char *) path);
    ufinfo = P_FINFO(curproc);

    for (i = 0; suid && ufinfo && i < 3; i++) {
	if (!ufinfo->fi_list[i].uf_file) {
	    if (!copyinstr(path, k_path, 255, (size_t *) & ret))
		cmn_err(CE_WARN,
			"%s closed while executing suid/sgid binary %s from "
			"%s (pid %d, uid %d), fake opening %s.", STDIO(i),
			k_path, u.u_psargs, ddi_get_pid(), CRED()->cr_ruid,
			STDIO(i));
	    ufalloc(i);
	    fake_opened_fd[i] = B_TRUE;
	} else
	    fake_opened_fd[i] = B_FALSE;
    }

  skip:
    ret = syscalls[EXECVE].sc(path, argv, envp);

    for (i = 0; config.secstdfd && suid && ufinfo && i < 3; i++)
	if (fake_opened_fd[i]) {
	    cmn_err(CE_WARN, "Closing fake opened %s.", STDIO(i));
	    closeandsetf(i, NULL);
	}

    rw_exit(&config_rwlock);
    dec_refcnt();

    return ret;
}

#ifdef _64BIT
int64_t pap_execve_64(const char *path, char *const argv[],
		      char *const envp[])
{
    uf_info_t *ufinfo;
    short i;
    boolean_t suid = B_FALSE, fake_opened_fd[3];
    char k_path[256];
    int64_t ret;

    inc_refcnt();
    dcmn_err((CE_CONT, "execve() syscall.\n"));
    rw_enter(&config_rwlock, RW_READER);

    if (!config.secstdfd)
	goto skip;

    suid = suid_check((char *) path);
    ufinfo = P_FINFO(curproc);

    for (i = 0; suid && ufinfo && i < 3; i++) {
	if (!ufinfo->fi_list[i].uf_file) {
	    if (!copyinstr(path, k_path, 255, (size_t *) & ret))
		cmn_err(CE_WARN,
			"%s closed while executing suid/sgid binary %s from "
			"%s (pid %d, uid %d), fake opening %s.", STDIO(i),
			k_path, u.u_psargs, ddi_get_pid(), CRED()->cr_ruid,
			STDIO(i));
	    ufalloc(i);
	    fake_opened_fd[i] = B_TRUE;
	} else
	    fake_opened_fd[i] = B_FALSE;
    }

  skip:
    ret = syscalls64[EXECVE].sc(path, argv, envp);

    for (i = 0; config.secstdfd && suid && ufinfo && i < 3; i++)
	if (fake_opened_fd[i]) {
	    cmn_err(CE_WARN, "Closing fake opened %s.", STDIO(i));
	    closeandsetf(i, NULL);
	}

    rw_exit(&config_rwlock);
    dec_refcnt();

    return ret;
}
#endif				/* _64 BIT */

#endif				/* SECSTDFD */
